#!/usr/bin/env python

from sympy import symbols
from sympy.physics.mechanics import dynamicsymbols, ReferenceFrame, Point

# Orientations
# ============

# right and left angles
r_theta_hip, r_theta_knee, r_theta_ankle = dynamicsymbols(
    "theta_hip_R, theta_knee_R, theta_ankle_R"
)
l_theta_hip, l_theta_knee, l_theta_ankle = dynamicsymbols(
    "theta_hip_L, theta_knee_L, theta_ankle_L"
)
theta_pelvis = dynamicsymbols("theta_pelvis")

ground_frame = ReferenceFrame("G")

# torso orientation
torso_frame = ReferenceFrame("H_ref")  # torso or head frame as name
torso_frame.orient(ground_frame, "Axis", (theta_pelvis, ground_frame.z))

# right foot orientations
right_thigh_frame = ReferenceFrame("Tr")
right_thigh_frame.orient(torso_frame, "Axis", (r_theta_hip, torso_frame.z))

right_shank_frame = ReferenceFrame("Sr")
right_shank_frame.orient(right_thigh_frame, "Axis", (r_theta_knee, right_thigh_frame.z))

right_foot_frame = ReferenceFrame("Fr")
right_foot_frame.orient(right_shank_frame, "Axis", (r_theta_ankle, right_shank_frame.z))

# left foot orientations
left_thigh_frame = ReferenceFrame("Tl")
left_thigh_frame.orient(torso_frame, "Axis", (l_theta_hip, torso_frame.z))

left_shank_frame = ReferenceFrame("Sl")
left_shank_frame.orient(left_thigh_frame, "Axis", (l_theta_knee, left_thigh_frame.z))

left_foot_frame = ReferenceFrame("Fl")
left_foot_frame.orient(left_shank_frame, "Axis", (l_theta_ankle, left_shank_frame.z))

# Point Locations
# ===============

O = Point("O")  # fixed point in the space
O.set_vel(ground_frame, 0)

x, y = dynamicsymbols("x, y")
hip_vel_x, hip_vel_y = dynamicsymbols("v_Hx, v_Hy")

# Joints
# ------

thigh_length, shank_length, foot_length = symbols("l_T, l_S, l_F")

hip = O.locatenew("H", x * ground_frame.x + y * ground_frame.y)  # Point('H')

# right knee and angle points
right_knee = Point("Kr")
# TODO Expand to 2d
right_knee.set_pos(hip, thigh_length * -right_thigh_frame.y)  # should it be - ??

right_ankle = Point("Ar")
right_ankle.set_pos(right_knee, shank_length * -right_shank_frame.y)

# left knee and angle points
left_knee = Point("Kl")
left_knee.set_pos(hip, thigh_length * -left_thigh_frame.y)  # should it be - ??

left_ankle = Point("Al")
left_ankle.set_pos(left_knee, shank_length * -left_shank_frame.y)


# Center of mass locations
# ------------------------

thigh_com_dist, shank_com_dist, foot_com_dist, torso_com_dist = symbols(
    "d_T, d_S, d_F, d_H"
)

# right center of mass locations
right_thigh_mass_center = Point("T_or")
right_thigh_mass_center.set_pos(hip, thigh_com_dist * -right_thigh_frame.y)

right_shank_mass_center = Point("S_or")
right_shank_mass_center.set_pos(right_knee, shank_com_dist * -right_shank_frame.y)

right_foot_mass_center = Point("F_or")
right_foot_mass_center.set_pos(
    right_ankle,
    (foot_com_dist - 0.06) * right_foot_frame.x + -0.035 * right_foot_frame.y,
)

# left center of mass locations
left_thigh_mass_center = Point("T_ol")
left_thigh_mass_center.set_pos(hip, thigh_com_dist * -left_thigh_frame.y)

left_shank_mass_center = Point("S_ol")
left_shank_mass_center.set_pos(left_knee, shank_com_dist * -left_shank_frame.y)

left_foot_mass_center = Point("F_ol")
left_foot_mass_center.set_pos(
    left_ankle, (foot_com_dist - 0.06) * left_foot_frame.x + -0.035 * left_foot_frame.y
)

torso_mass_center = Point("H_o")
torso_mass_center.set_pos(hip, torso_com_dist * torso_frame.y)

# Define kinematical differential equations
# =========================================

# right and left angles
r_omega_hip, r_omega_knee, r_omega_ankle = dynamicsymbols(
    "omega_hip_R, omega_knee_R, omega_ankle_R"
)
l_omega_hip, l_omega_knee, l_omega_ankle = dynamicsymbols(
    "omega_hip_L, omega_knee_L, omega_ankle_L"
)
omega_torso = dynamicsymbols("omega_torso")

time = symbols("t")

kinematical_differential_equations = [
    hip_vel_x - x.diff(),
    hip_vel_y - y.diff(),
    omega_torso - theta_pelvis.diff(),
    r_omega_hip - r_theta_hip.diff(),
    r_omega_knee - r_theta_knee.diff(),
    r_omega_ankle - r_theta_ankle.diff(),
    l_omega_hip - l_theta_hip.diff(),
    l_omega_knee - l_theta_knee.diff(),
    l_omega_ankle - l_theta_ankle.diff(),
]


# Angular Velocities
# ==================

# torso frame angular velocity
torso_frame.set_ang_vel(ground_frame, omega_torso * ground_frame.z)

# right frame angular velocities
right_thigh_frame.set_ang_vel(torso_frame, r_omega_hip * torso_frame.z)
right_shank_frame.set_ang_vel(right_thigh_frame, r_omega_knee * right_thigh_frame.z)
right_foot_frame.set_ang_vel(right_shank_frame, r_omega_ankle * right_shank_frame.z)

# left frame angular velocities
left_thigh_frame.set_ang_vel(torso_frame, l_omega_hip * torso_frame.z)
left_shank_frame.set_ang_vel(left_thigh_frame, l_omega_knee * left_thigh_frame.z)
left_foot_frame.set_ang_vel(left_shank_frame, l_omega_ankle * left_shank_frame.z)


# Linear Velocities
# =================

hip.set_vel(ground_frame, hip_vel_x * ground_frame.x + hip_vel_y * ground_frame.y)

# torso velocity
torso_mass_center.v2pt_theory(hip, ground_frame, torso_frame)

# right leg velocities
right_thigh_mass_center.v2pt_theory(hip, ground_frame, right_thigh_frame)

right_knee.v2pt_theory(hip, ground_frame, right_thigh_frame)
right_shank_mass_center.v2pt_theory(right_knee, ground_frame, right_shank_frame)

right_ankle.v2pt_theory(right_knee, ground_frame, right_shank_frame)
right_foot_mass_center.v2pt_theory(right_ankle, ground_frame, right_foot_frame)

# left leg velocities
left_thigh_mass_center.v2pt_theory(hip, ground_frame, left_thigh_frame)

left_knee.v2pt_theory(hip, ground_frame, left_thigh_frame)
left_shank_mass_center.v2pt_theory(left_knee, ground_frame, left_shank_frame)

left_ankle.v2pt_theory(left_knee, ground_frame, left_shank_frame)
left_foot_mass_center.v2pt_theory(left_ankle, ground_frame, left_foot_frame)